module Board where

import System.Random 
import Data.List     (elemIndices, transpose)
import Control.Parallel.Strategies
import Control.DeepSeq

type Board = [[Int]]

data Direction = U | D | L | R
  deriving (Enum, Bounded)

instance NFData Direction where
    rnf _ = ()

instance Show Direction where
    show U = "Up"
    show D = "Down"
    show L = "Left"
    show R = "Right"

instance Read Direction where
    readsPrec _ value = 
        case value of
            "Up"    -> [(U, "")]
            "Down"  -> [(D, "")]
            "Left"  -> [(L, "")]
            "Right" -> [(R, "")]
            _       -> []

instance Eq Direction where
    U == U = True
    D == D = True
    L == L = True
    R == R = True
    _ == _ = False

instance Ord Direction where
    compare U U = EQ
    compare U _ = LT
    compare _ U = GT
    compare D D = EQ
    compare D _ = LT
    compare _ D = GT
    compare L L = EQ
    compare L _ = LT
    compare _ L = GT
    compare R R = EQ

moveLeft :: Board -> Board
moveLeft b = map slideRowLeft b where 
        slideRowLeft [] = []
        slideRowLeft [x] = [x]
        slideRowLeft (x:y:zs)
          | x == 0 = slideRowLeft (y : zs) ++ [0]
          | y == 0 = slideRowLeft (x : zs) ++ [0] 
          | x == y = (x + y) : slideRowLeft zs ++ [0]
          | otherwise = x : slideRowLeft (y : zs)

move :: Direction -> Board -> Board
move U b = transpose . moveLeft . transpose $ b
move D b = transpose . map reverse . moveLeft . map reverse . transpose $ b
move L b = moveLeft b
move R b = map reverse . moveLeft . map reverse $ b

getEmptyTiles :: Board -> [(Int, Int)] 
getEmptyTiles b = concatMap (\(n, row) -> zip (replicate 4 n) (elemIndices 0 row)) (zip [0..3] b)

randomTileValue :: IO Int
randomTileValue = do
  x <- randomRIO (1, 10 :: Int) --assumes 90/10 2/4 probability
  return $ if x == 1 then 4 else 2

selectRandomTile :: [(Int, Int)] -> IO (Maybe (Int, Int))
selectRandomTile [] = return Nothing
selectRandomTile tiles = do
  i <- randomRIO (0, length tiles - 1)
  return $ Just (tiles !! i)

updateTile :: (Int, Int) -> Int -> Board -> Board
updateTile (r, c) val b = updateIndex (updateIndex (const val) c) r b where
    updateIndex fn i list = take i list ++ fn (head $ drop i list) : tail (drop i list)

addTile :: Board -> IO Board
addTile b = do
  let emptyTiles = getEmptyTiles b
  newValue <- randomTileValue
  maybePoint <- selectRandomTile emptyTiles
  case maybePoint of
    Just newPoint -> return $ updateTile newPoint newValue b
    Nothing -> return b -- No more empty tiles, return the board as is

getAvailableMoves :: Board -> [Board]
getAvailableMoves b = map (\x -> move x b) $ filter (\dir -> validMove dir b) moves
  where moves = [U,D,L,R]

validMove :: Direction -> Board -> Bool
validMove dir b = move dir b /= b

-- PRINT
printBoard :: Board -> IO ()
printBoard b = do
    putStrLn "-------------"
    putStrLn $ show (b !! 0)
    putStrLn $ show (b !! 1)
    putStrLn $ show (b !! 2)
    putStrLn $ show (b !! 3)
    putStrLn "-------------"