module Main where

import System.Random
import Control.Monad (when)
import Board
import Minimax
import Data.Time.Clock
import Data.Csv
import qualified Data.ByteString.Lazy as BL
import qualified Data.Vector as V

data GameState = GameState {
    board :: Board,
    isPlayerTurn :: Bool,
    gameOver :: Bool
}

-- Initialize the game state with two random tiles
initGame :: IO GameState
initGame = do
    initialBoard <- addTile =<< addTile (replicate 4 (replicate 4 0))
    return $ GameState initialBoard True False

-- Check if the game is over
checkGameOver :: Board -> Bool
checkGameOver b = null (getEmptyTiles b) && not (canMove b)

doublesToCsv :: [Double] -> BL.ByteString
doublesToCsv doubles = encode $ map (\x -> [x]) doubles

data TimeRecord = TimeRecord { timeTaken :: Double }

instance ToRecord TimeRecord where
    toRecord (TimeRecord t) = record [toField t]

gameLoop :: GameState -> [Double] -> IO ()
gameLoop state d
    | gameOver state = do
        putStrLn $ "Game Over!"
        putStrLn $ "Score: " ++ show (maximum (map maximum (board state)))
        BL.writeFile "output6-2.csv" $ doublesToCsv d
    | otherwise = do
        printBoard (board state)
        if isPlayerTurn state 
            then do
                startTime <- getCurrentTime
                newState <- playerTurn state
                endTime <- getCurrentTime
                let timeTaken = realToFrac $ (diffUTCTime endTime startTime) :: Double
                gameLoop newState $ d ++ [timeTaken]
            else do
                newState <- computerTurn state
                gameLoop newState $ d

-- Handle player's turn
playerTurn :: GameState -> IO GameState
playerTurn state = do
    let b = board state
    moveDir <- bestMove b -- `Maybe Direction`
    case moveDir of
        Just dir -> do
            let newBoard = move dir b
            return state {board = newBoard, isPlayerTurn = False}
        Nothing -> 
            return state

-- Handle computer's turn by adding a random tile
computerTurn :: GameState -> IO GameState
computerTurn state = do
    newBoard <- addTile (board state)
    let newState = state {board = newBoard, isPlayerTurn = True}
    return newState {gameOver = checkGameOver newBoard}

-- Start the game
main :: IO ()
main = do
    a <- initGame
    gameLoop a []
